// Copyright © Microsoft Corporation
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.opengroup.osdu.register.provider.azure.action;

import java.util.List;
import java.util.stream.Collectors;

import org.opengroup.osdu.azure.cosmosdb.CosmosStore;
import org.opengroup.osdu.core.common.logging.JaxRsDpsLog;
import org.opengroup.osdu.core.common.model.http.AppException;
import org.opengroup.osdu.core.common.model.http.DpsHeaders;
import org.opengroup.osdu.register.action.model.Action;
import org.opengroup.osdu.register.provider.azure.di.AzureBootstrapConfig;
import org.opengroup.osdu.register.provider.azure.di.CosmosContainerConfig;
import org.opengroup.osdu.register.provider.interfaces.action.IActionRepo;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Repository;

import com.azure.cosmos.models.CosmosQueryRequestOptions;
import com.azure.cosmos.models.SqlParameter;
import com.azure.cosmos.models.SqlQuerySpec;

@Repository
public class ActionRepository implements IActionRepo {

    @Autowired
    private AzureBootstrapConfig azureBootstrapConfig;

    @Autowired
    private CosmosContainerConfig cosmosContainerConfig;

    @Autowired
    private CosmosStore cosmosStore;

    @Autowired
    private DpsHeaders dpsHeaders;

    @Autowired
    private JaxRsDpsLog logger;

    @Override
    public Action createAction(Action action) {

        if(action.getId() == null){
            logger.error("Action id cannot be null");
            throw new AppException(400, "Bad Request", "Action id cannot be null");
        }

        ActionDoc doc = new ActionDoc(action, dpsHeaders.getPartitionId());

        try {
            cosmosStore.createItem(dpsHeaders.getPartitionId(), azureBootstrapConfig.getCosmosDBName(), cosmosContainerConfig.getActionContainerName(), dpsHeaders.getPartitionId(), doc);
        }
        catch (AppException e) {
            if(e.getError().getCode() == 409) {
                logger.error(String.format("An action already exists with the id: %s", action.getId()));
                throw new AppException(409, "Conflict", String.format("An action already exists with the id: %s", action.getId()));
            }
            else {
                logger.error(e.getMessage());
                throw new AppException(e.getError().getCode(), e.getError().getReason(), e.getMessage());
            }
        }

        return action;
    }

    @Override
    public Action get(String id) {

        ActionDoc doc = cosmosStore.findItem(dpsHeaders.getPartitionId(), azureBootstrapConfig.getCosmosDBName(), cosmosContainerConfig.getActionContainerName(), id, dpsHeaders.getPartitionId(), ActionDoc.class)
                .orElseThrow(() -> {
                    logger.error(String.format("Action with id %s does not exist.", id));
                    return new AppException(404, "Not found", String.format("Action with id %s does not exist.", id));
                });

        return convertToActionClass(doc);
    }

    @Override
    public boolean delete(String id) {
        try {
            cosmosStore.deleteItem(dpsHeaders.getPartitionId(), azureBootstrapConfig.getCosmosDBName(), cosmosContainerConfig.getActionContainerName(), id, dpsHeaders.getPartitionId());
        }
        catch (AppException e) {
            if(e.getError().getCode() == 404) {
                logger.error(String.format("Action with id %s does not exist.", id));
            }
            else {
                logger.error(e.getMessage());
                throw new AppException(e.getError().getCode(), e.getError().getReason(), e.getMessage());
            }
            return false;
        }
        return true;
    }

    @Override
    public List<Action> getAllActions() {
        SqlQuerySpec query = new SqlQuerySpec("SELECT * FROM c where c.dataPartitionId = @partitionId");

        query.getParameters().add(new SqlParameter("@partitionId", dpsHeaders.getPartitionId()));

        CosmosQueryRequestOptions options = new CosmosQueryRequestOptions();
        return cosmosStore.queryItems(dpsHeaders.getPartitionId(), azureBootstrapConfig.getCosmosDBName(), cosmosContainerConfig.getActionContainerName(), query, options, ActionDoc.class)
                .stream().map(this::convertToActionClass).collect(Collectors.toList());
    }

    private Action convertToActionClass(ActionDoc doc){
        Action action = new Action();
        action.setId(doc.getId());
        action.setName(doc.getName());
        action.setDescription(doc.getDescription());
        action.setImg(doc.getImg());
        action.setUrl(doc.getUrl());
        action.setContactEmail(doc.getContactEmail());
        action.setCreatedOnEpoch(doc.getCreatedOnEpoch());
        action.setFilter(doc.getFilter());
        return action;
    }

}
