// Copyright © Microsoft Corporation
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.opengroup.osdu.register.provider.azure.subscriber;

import java.util.List;

import org.opengroup.osdu.core.common.logging.JaxRsDpsLog;
import org.opengroup.osdu.core.common.model.http.AppException;
import org.opengroup.osdu.register.provider.azure.subscriber.pubsub.IPubsubSubscription;
import org.opengroup.osdu.register.provider.interfaces.subscriber.ISubscriptionRepository;
import org.opengroup.osdu.register.subscriber.model.Secret;
import org.opengroup.osdu.register.subscriber.model.Subscription;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Repository;

@Repository
public class SubscriptionRepository implements ISubscriptionRepository {

    @Autowired
    private IPubsubSubscription pubSubSubscription;

    @Autowired
    private CosmosSubscription cosmosSubscription;

    @Autowired
    private JaxRsDpsLog logger;

    /**
     * @param input Subscription Object
     * @return Subscription Object
     * <p>
     * This method creates a Record for the subscription in CosmosDb along
     * with a Push Subscription in Event Grid.
     * <p>
     * Expected Flow -> Create a record in Cosmos Db along with a Push Subscription in Event Grid
     * Possible Failures ->
     * 1) Creating PushSubscription in Event Grid
     * How it is handled -> Delete the record in Cosmos Db as well
     * Possible Issue -> What happens if this delete operation on Cosmos Db fails?
     * Users won't be able to create a subscription again with same Topic and
     * push endpoint combination since system will generate 409 on next retry
     * because of presence of record with same Id in Cosmos Db
     * <p>
     * How to resolve this?
     * createPushSubscriptionIfDoesNotExist method takes care of that
     * When Cosmos Db throws 409 while creating a new record we can check if a Push Subscription
     * for that Record already exists in Event Grid or not.
     * If present -> Return 409
     * If not present -> We can try creating the Push Subscription again in Event Grid
     * Also we will update the record in Cosmos Db in case other fields might
     * be different like description etc.
     */

    @Override
    public Subscription create(Subscription input) {
        Subscription subscription = null;
        try {
            subscription = cosmosSubscription.createSubscription(input);
        } catch (AppException e) {
            if (e.getError().getCode() == 409) {
                // This flow is added to handle the scenario where creating PubSub Subscription in EventGrid failed
                // and deleting the corresponding record from Cosmos Db was also unsuccessful.
                // This will result in an 500 Exception so the user should be able to create the subscription with
                // the same topic and PubSub Endpoint combination again
                logger.info(String.format("Record already exists for Subscription with id %s", input.getId()));
                Subscription outputSubscription = createPubSubSubscriptionIfDoesNotExist(subscription);
                input.setNotificationId(outputSubscription.getNotificationId());
                return input;
            } else {
                logger.error(e.getMessage());
                throw new AppException(e.getError().getCode(), e.getError().getReason(), e.getMessage());
            }
        }
        try {
            pubSubSubscription.createSubscription(subscription.getNotificationId(), subscription.getTopic());
            logger.debug("PubSub Subscription created with ID:" + subscription.getNotificationId());
            return input;
        } catch (AppException e) {
            if (e.getError().getCode() == 409) {
                throw new AppException(409, "Conflict", "Another request is trying to create the same subscription");
            }
            logger.error("PubSub Subscription creation failed.Rollback Cosmos Subscription Creation");
            try {
                cosmosSubscription.deleteSubscription(input.getId());
            } catch (AppException ignored) {
                // TODO:HANDLE THIS
            }
            logger.error("Unexpected error creating subscription");
            throw new AppException(500, "Server Error", "Unexpected error creating subscription");
        }
    }

    @Override
    public Subscription get(String id) {
        return cosmosSubscription.getSubscription(id);
    }

    @Override
    public List<Subscription> query(String notificationId) {
        return cosmosSubscription.querySubscription(notificationId);
    }

    /**
     * @param id Identifier for the subscription
     * @return true if subscription successfully deleted
     * false if subscription not found
     * Exception thrown -> AppException with Status Codes 409 and 500
     */
    @Override
    public boolean delete(String id) {
        Subscription subscription = cosmosSubscription.getSubscription(id);
        if (subscription != null) {
            logger.info("Record not found for subscription with ID: " + id);
            return false;
        }
        logger.debug("Record found for subscription with ID: " + id);
        try {
            pubSubSubscription.deleteSubscription(subscription.getNotificationId(), subscription.getTopic());
            logger.debug("PubSub Subscription deleted with ID:" + subscription.getNotificationId());
            try {
                cosmosSubscription.deleteSubscription(id);
                return true;
            } catch (AppException e) {
                if (e.getError().getCode() == 404) {
                    logger.info(String.format("Subscriber with id %s does not exist.", id));
                    return false;
                } else {
                    throw e;
                }
            }
        } catch (AppException e) {
            if (e.getError().getCode() == 409) {
                logger.error("Another request is trying to delete the same subscription");
                throw new AppException(409, "Conflict", "Another request is trying to delete the same subscription");
            } else {
                logger.error("Unexpected error deleting subscription");
                throw new AppException(500, "Server Error", "Unexpected error deleting subscription");
            }
        }
    }

    @Override
    public boolean patch(Subscription subscription, Secret secret) {
        return cosmosSubscription.updateSecret(subscription, secret);
    }

    @Override
    public List<Subscription> getAll() {
        return cosmosSubscription.getAllSubscriptions();
    }

    private Subscription createPubSubSubscriptionIfDoesNotExist(Subscription input) {
        try {
            // We are fetching the record from cosmos db again because we want to use the original
            // notification-id to create the PubSub Subscription
            Subscription originalSubscription = cosmosSubscription.getSubscription(input.getId());
            if (originalSubscription != null)
                input.setNotificationId(originalSubscription.getNotificationId());
            logger.debug(String.format("Creating PubSub Subscription with id %s if does not exists", input.getNotificationId()));

            // We will check if PubSub Subscription does not exist then we should try creating it again since
            // the corresponding record in the Cosmos Db is already present
            pubSubSubscription.checkIfSubscriptionExists(input.getNotificationId(), input.getTopic());
            logger.error("A subscriber already exists with the same topic and endpoint combination");
            throw new AppException(409, "Conflict", "A subscriber already exists with the same topic and endpoint combination");
        } catch (AppException e) {
            if (e.getError().getCode() == 409) {
                throw e;
            } else if (e.getError().getCode() == 404) {
                try {
                    // If PubSub Subscription is not yet created we should create it, also we might
                    // need to update the cosmos db record since some other fields might be different
                    cosmosSubscription.updateSubscription(input);
                    // Now creating the PubSub Subscription, if it fails again we should throw server error
                    // otherwise we can return Success code as response
                    pubSubSubscription.createSubscription(input.getNotificationId(), input.getTopic());
                    logger.info("PubSub Subscription created with ID's:" + input.getNotificationId());
                    return input;
                } catch (AppException exception) {
                    if (exception.getError().getCode() == 409) {
                        logger.error("Another request is trying to create the same PubSub subscription");
                        throw new AppException(409, "Conflict", "Another request is trying to create the same PubSub subscription");
                    } else {
                        logger.error("Unexpected error creating subscription");
                        throw new AppException(500, "Server Error", "Unexpected error creating subscription");
                    }
                }
            } else {
                logger.error("Unexpected error creating subscription");
                throw new AppException(500, "Server Error", "Unexpected error creating subscription");
            }
        }
    }

}
