// Copyright © Microsoft Corporation
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.opengroup.osdu.register.provider.azure.subscriber.pubsub;

import com.microsoft.azure.CloudException;
import org.opengroup.osdu.azure.partition.PartitionServiceClient;
import org.opengroup.osdu.core.common.logging.JaxRsDpsLog;
import org.opengroup.osdu.core.common.model.http.AppException;
import org.opengroup.osdu.core.common.model.http.DpsHeaders;
import org.opengroup.osdu.register.provider.azure.subscriber.messageBus.IManagementClientFactory;
import org.opengroup.osdu.register.provider.azure.subscriber.TopicsRepositoryImpl;
import org.opengroup.osdu.register.provider.azure.util.PartitionHelper;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.autoconfigure.condition.ConditionalOnProperty;
import org.springframework.stereotype.Component;

@Component
@ConditionalOnProperty(value = "azure.servicebus.enabled", havingValue = "true", matchIfMissing = false)
public class PullSubscription implements IPubsubSubscription {
    // TODO: UPDATE ERROR_MESSAGE FOR SERVICE BUS
    private static final String RESOURCE_PROVISIONING_ERROR_MESSAGE = "Resource cannot be updated during provisioning";
    @Autowired
    private IManagementClientFactory factory;
    @Autowired
    private DpsHeaders dpsHeaders;
    @Autowired
    private JaxRsDpsLog logger;
    @Autowired
    private PartitionServiceClient partitionService;
    @Autowired
    private PartitionHelper partitionHelper;

    public void createSubscription(String subscriptionId, String topicName) {
        try {
            TopicsRepositoryImpl topicsRepositoryImpl = new TopicsRepositoryImpl(logger, dpsHeaders, partitionHelper);
            String topicAliasName = topicsRepositoryImpl.getTopicAliasFromFullyQualifiedTopicName(topicName);
            factory.getManager().createSubscription(topicAliasName, subscriptionId);
        } catch (Exception e) {
            if (e.getMessage().equals(RESOURCE_PROVISIONING_ERROR_MESSAGE)) {
                logger.error("Another request is trying to create the same Pull subscription");
                throw new AppException(409, "Conflict", "Another request is trying to create the same Push subscription");
            } else {
                logger.error("Creating Pull Subscription failed with error: " + e.toString());
                throw new AppException(500, "Server Error", "Unexpected error creating Push subscription");
            }
        }
    }

    public void deleteSubscription(String subscriptionId, String topicName) {
        try {
            TopicsRepositoryImpl topicsRepositoryImpl = new TopicsRepositoryImpl(logger, dpsHeaders, partitionHelper);
            String topicAliasName = topicsRepositoryImpl.getTopicAliasFromFullyQualifiedTopicName(topicName);
            factory.getManager().deleteSubscription(topicAliasName, subscriptionId);
        } catch (Exception e) {
            if (e instanceof CloudException) {
                CloudException cloudException = (CloudException) e;
                logger.error(cloudException.toString());
                throw new AppException(cloudException.response().code(), cloudException.body().code(), cloudException.body().message());
            } else {
                logger.error("Deleting Pull Subscription failed with error: " + e.toString());
                throw new AppException(500, "Server Error", "Unexpected error deleting Pull subscription");
            }
        }
    }

    public void checkIfSubscriptionExists(String subscriptionId, String topicName) {
        try {
            TopicsRepositoryImpl topicsRepositoryImpl = new TopicsRepositoryImpl(logger, dpsHeaders, partitionHelper);
            String topicAliasName = topicsRepositoryImpl.getTopicAliasFromFullyQualifiedTopicName(topicName);
            factory.getManager().getSubscription(topicAliasName, subscriptionId);
        } catch (Exception e) {
            if (e instanceof NullPointerException) {
                logger.error(String.format("Pull Subscription with id %s does not exist.", subscriptionId));
                throw new AppException(404, "Not found", String.format("Pull Subscription with id %s does not exist.", subscriptionId));
            } else {
                logger.error("Fetching Pull subscription failed with error: " + e.toString());
                throw new AppException(500, "Server Error", "Unexpected error while Fetching Pull subscription");
            }
        }
    }
}
