#!/bin/bash

# Exit as soon as a command fails
set -e

{{$definitionName := getenv "BUILD_DEFINITIONNAME" -}}
{{$sourceBranchName := getenv "BUILD_SOURCEBRANCHNAME" -}}
{{$buildNumber := getenv "BUILD_BUILDNUMBER" -}}

SUBSCRIBER_SECRET=$1
SUBSCRIBER_PRIVATE_KEY_ID=$2

if [ -z "$SUBSCRIBER_SECRET" ]
then
  echo "SUBSCRIBER_SECRET can not be empty"
  exit 1
fi

BUILD_DEFINITIONNAME={{$definitionName}}
BUILD_SOURCEBRANCHNAME={{$sourceBranchName}}
BUILD_BUILDNUMBER={{$buildNumber}}
SCRIPT_DIR=$(dirname $0)

echo "This script will deploy the app to $GCLOUD_PROJECT"

if [ "$GCLOUD_PROJECT" = "" ]
then
    export GCLOUD_PROJECT=$(gcloud config get-value project)
fi

echo "  Environment = $ENVIRONMENT"
echo "  Project = $GCLOUD_PROJECT"
echo "  Build definition = $BUILD_DEFINITIONNAME"
echo "  Source branch = $BUILD_SOURCEBRANCHNAME"
echo "  Build number = $BUILD_BUILDNUMBER"
echo "  Directory = $SCRIPT_DIR"

echo "Print Directory contents"
ls $SCRIPT_DIR

VERSION_NUMBER=$BUILD_BUILDNUMBER
VERSION_NUMBER=$(echo "$VERSION_NUMBER" | tr _ - | tr '[:upper:]' '[:lower:]')
## UGLY.. removes '.' from string.
VERSION_NUMBER=${VERSION_NUMBER//.}
SERVICE_NAME="os-register"

count=$(gcloud app services list --project $GCLOUD_PROJECT | grep $SERVICE_NAME | wc -l)
if [ $count -gt 0 ]; then
  CURRENT_VERSION=$(gcloud app services describe $SERVICE_NAME --project $GCLOUD_PROJECT --format=json | jq --raw-output '.split.allocations | keys[0]')
else
  CURRENT_VERSION=""
fi
echo "Current version = $CURRENT_VERSION"
echo "Version to be deployed = $VERSION_NUMBER"

if [ "$VERSION_NUMBER" != "$CURRENT_VERSION" -o "$BUILD_FORCE_DEPLOY" = "true" -o "$BUILD_FORCE_DEPLOY" = "True" -o "$BUILD_FORCE_DEPLOY" = "TRUE" ]
then
    if [ "$BUILD_SOURCEBRANCHNAME" = "master" -o "$BUILD_FORCE_DEPLOY" = "true" -o "$BUILD_FORCE_DEPLOY" = "True" -o "$BUILD_FORCE_DEPLOY" = "TRUE" ]
    then
        APPYAML=app.yaml

        echo "apply sed on app yaml"
        sed -i -e "s|GOOGLE-AUDIENCES|$GOOGLE_AUDIENCES|g" $APPYAML
        sed -i -e "s|SUBSCRIBER-SECRET|$SUBSCRIBER_SECRET|g" $APPYAML
        sed -i -e "s|SUBSCRIBER-PRIVATE-KEY-ID|$SUBSCRIBER_PRIVATE_KEY_ID|g" $APPYAML

        echo "print app yaml after sed"
        cat $APPYAML

        # If the version $VERSION_NUMBER already exists, but not serving traffic, gcloud app deploy will fail. In this case, simply start that version and migrate traffic to it
        doesVersionExist=`gcloud app versions list --service=$SERVICE_NAME --project=$GCLOUD_PROJECT | grep $VERSION_NUMBER | wc -l`
        if [ "$doesVersionExist" -eq "1" ]
        then
            echo "Starting version $VERSION_NUMBER"
            gcloud app versions start $VERSION_NUMBER --service=$SERVICE_NAME --project=$GCLOUD_PROJECT --quiet
            echo "Migrating all traffic to version $VERSION_NUMBER"
            gcloud app services set-traffic $SERVICE_NAME --splits $VERSION_NUMBER=1 --project=$GCLOUD_PROJECT --quiet
        else
            echo "Deploying service to Google project $GCLOUD_PROJECT"
            gcloud app deploy --quiet --version=$VERSION_NUMBER --project=$GCLOUD_PROJECT $APPYAML --log-http --verbosity=debug
            echo "Service deployed successfully"

        fi
    else
        echo "Not deploying the application because build source branch was not master"
    fi
else
    echo "Not deploying the application because $VERSION_NUMBER is already deployed and force deploy flag is not set to true"
fi

# disable exit on error
set +e
echo "Creating service account identity to used by service"
result=$(gcloud beta iam service-accounts create de-notification-service --project $GCLOUD_PROJECT --description "Service account used on push notifications sent from the Data Ecosystem" --display-name "DE notification-service account" 2>&1)

echo $result

if echo $result | grep -q -e "Service account de-notification-service already exists" -e "Created service account \[de-notification-service\]"; then
    echo "Successfully created or service account already exists."
    gcloud projects add-iam-policy-binding "$GCLOUD_PROJECT" --member "serviceAccount:de-notification-service@$GCLOUD_PROJECT.iam.gserviceaccount.com" --role roles/iam.serviceAccountUser
else
    echo "Unexpected error creating service account."
    exit 2
fi